package perfSONAR_PS::Services::LS;
=head1 NAME

perfSONAR_PS::Services::LS - This is a stub module to make it easier for
installing the Lookup Service. This allows them install
"perfSONAR_PS::Services::LS" instead of having to install
"perfSONAR_PS::Services::LS::LS"

=cut

1;
